/**
 * \file src/gui/xic_view/xic_box/xicbox.cpp
 * \date 12/1/2018
 * \author Olivier Langella
 * \brief XIC box widget
 */
/*******************************************************************************
* Copyright (c) 2018 Olivier Langella <olivier.langella@u-psud.fr>.
*
* This file is part of XTPcpp.
*
*     XTPcpp is free software: you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation, either version 3 of the License, or
*     (at your option) any later version.
*
*     XTPcpp is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with XTPcpp.  If not, see <http://www.gnu.org/licenses/>.
*
* Contributors:
*     Olivier Langella <olivier.langella@u-psud.fr> - initial API and implementation
******************************************************************************/

#include "xicbox.h"
#include "ui_xic_box.h"
#include <QMessageBox>
#include <pappsomspp/exception/exceptionnotfound.h>
#include "../xicworkerthread.h"
#include "../../project_view/projectwindow.h"
#include <pappsomspp/xic/detection/xicdetectionzivy.h>
#include <pappsomspp/exception/exceptionnotpossible.h>



class XicDetectionList: public pappso::XicDetectionSinkInterface {
public:
    XicDetectionList () {
        _max_peak.setArea(0);
    }
    void setPeptideEvidenceList(const std::vector<const PeptideEvidence *> & peptide_evidence_list) {
        for (const PeptideEvidence * p_evidence: peptide_evidence_list) {
            _rt_list.push_back(p_evidence->getRetentionTime());
        }
    }
    void setXicPeak (pappso::XicPeak & xic_peak) override {
        _peak_list.push_back(xic_peak);

        for (pappso::pappso_double rt: _rt_list) {
            if (xic_peak.containsRt(rt)) {
                if (xic_peak.getArea() > _max_peak.getArea()) {
                    _max_peak = xic_peak;
                }
            }
        }
    };
    const pappso::XicPeak & getMatchedPeak() const {
        return _max_peak;
    }
    const std::vector<pappso::XicPeak> & getXicPeakList() const {
        return _peak_list;
    };
    void clear() {
        _peak_list.clear();
        _max_peak.setArea(0);
    };
private :
    unsigned int _count=0;
    std::vector<pappso::XicPeak> _peak_list;
    std::vector<pappso::pappso_double> _rt_list;
    pappso::XicPeak _max_peak;
};



XicBox::XicBox(XicWindow *parent):
    QWidget(parent),
    ui(new Ui::XicBox)
{
    _p_xic_window = parent;
    ui->setupUi(this);

    XicWorkerThread * p_worker = new XicWorkerThread();
    p_worker->moveToThread(&_xic_thread);
    _xic_thread.start();


    //ui->histo_widget->setInteractions(QCP::iRangeDrag | QCP::iRangeZoom );
    //ui->xic_list_widget->setLayout(new QVBoxLayout(ui->xic_list_widget));
    ui->histo_widget->legend->setVisible(true);

#if QT_VERSION >= 0x050000
    // Qt5 code
    connect(this, &XicBox::loadXic, p_worker, &XicWorkerThread::doXicLoad);
    connect(p_worker, &XicWorkerThread::xicLoaded,this, &XicBox::setXic);
    connect(p_worker, &XicWorkerThread::operationFailed, this, &XicBox::error);


    connect(this, &XicBox::computeIsotopeMassList, p_worker, &XicWorkerThread::doComputeIsotopeMassList);
    connect(p_worker, &XicWorkerThread::isotopeMassListComputed,this, &XicBox::setIsotopeMassList);

    connect(_p_xic_window, &XicWindow::reExtractXicNeeded,this, &XicBox::reExtractXic);

#else
// Qt4 code
#endif


    QStringList msrun_list;
    for (MsRunSp msrun_sp : _p_xic_window->getProjectWindow()->getProjectP()->getMsRunStore().getMsRunList()) {
        msrun_list << msrun_sp.get()->getFilename();
        qDebug() << "ProteinListWindow::setIdentificationGroup " << msrun_sp.get()->getFilename();
    }
    QCompleter *completer = new QCompleter(msrun_list, this);
    completer->setCaseSensitivity(Qt::CaseInsensitive);

    completer->setCompletionMode(QCompleter::PopupCompletion);
    completer->setModelSorting(QCompleter::CaseSensitivelySortedModel);
    completer->setFilterMode(Qt::MatchContains);
    ui->msrun_autocompletion->setCompleter(completer);

}

XicBox::~XicBox()
{
    qDebug() << "XicBox::~XicBox";
    _xic_thread.quit();
    _xic_thread.wait();
}

void XicBox::remove() {
    _p_xic_window->removeXicBox(this);
}

void XicBox::error(QString error_message) {
    QMessageBox::warning(this,
                         tr("Error trying to save spectrum to SVG file :"), error_message);

}

void XicBox::extractXicInOtherMsRun() {
    qDebug() << "XicBox::extractXicInOtherMsRun begin";
    try {
        MsRunSp msrun_sp = _p_xic_window->getProjectWindow()->getProjectP()->getMsRunStore().getInstance(ui->msrun_autocompletion->text());
        _p_xic_window->addXicInMsRun(_p_peptide_evidence, msrun_sp);
    }
    catch (pappso::ExceptionNotFound & error) {
        qDebug() << "XicBox::extractXicInOtherMsRun not found " << ui->msrun_autocompletion->text();
    }
}

void XicBox::setPeptideEvidence(const PeptideEvidence * p_peptide_evidence) {


    _p_peptide_evidence = p_peptide_evidence;
    _msrun_sp = p_peptide_evidence->getIdentificationDataSource()->getMsRunSp();

    ui->peptide_label->setText(_p_peptide_evidence->getPeptideXtpSp().get()->toString());
    ui->msrun_label->setText(_p_peptide_evidence->getMsRunP()->getFilename());
    ui->charge_label->setText(QString("%1").arg(_p_peptide_evidence->getCharge()));
    ui->mz_label->setText(QString::number(_p_peptide_evidence->getTheoreticalMz(), 'f', 4));

    //get same xic peptide evidence (msrun, peptide, charge)
    //p_projet
    _peptide_evidence_list.clear();
    _p_xic_window->getProjectWindow()->getProjectP()->getSameXicPeptideEvidenceList(_peptide_evidence_list, _msrun_sp.get(),_p_peptide_evidence->getPeptideXtpSp().get(),_p_peptide_evidence->getCharge());

    emit computeIsotopeMassList(_p_peptide_evidence->getPeptideXtpSp(), _p_peptide_evidence->getCharge(), _p_xic_window->getXicExtractPrecision(), 0.8);

}

void XicBox::setPeptideEvidenceInMsRun(const PeptideEvidence * p_peptide_evidence, MsRunSp msrun_sp) {
    _p_peptide_evidence = p_peptide_evidence;
    _msrun_sp = msrun_sp;

    ui->peptide_label->setText(_p_peptide_evidence->getPeptideXtpSp().get()->toString());
    ui->msrun_label->setText(msrun_sp->getFilename());
    ui->charge_label->setText(QString("%1").arg(_p_peptide_evidence->getCharge()));
    ui->mz_label->setText(QString::number(_p_peptide_evidence->getTheoreticalMz(), 'f', 4));

    //get same xic peptide evidence (msrun, peptide, charge)
    //p_projet
    _peptide_evidence_list.clear();

    _p_xic_window->getProjectWindow()->getProjectP()->getSameXicPeptideEvidenceList(_peptide_evidence_list, _msrun_sp.get(),_p_peptide_evidence->getPeptideXtpSp().get(),_p_peptide_evidence->getCharge());
    //_p_xic_window->getProjectWindow()->getProjectP()->getSameXicPeptideEvidenceList(p_peptide_evidence, _peptide_evidence_list);

    emit computeIsotopeMassList(_p_peptide_evidence->getPeptideXtpSp(), _p_peptide_evidence->getCharge(), _p_xic_window->getXicExtractPrecision(), 0.8);

}

void XicBox::setXic(std::vector< pappso::XicSp> xic_sp_list) {
    qDebug() << "XicBox::setXic begin " << xic_sp_list.size();
    ui->xic_widget->clear();
    //pappso::XicWidget * xic_widget = new pappso::XicWidget(this);
    //ui->xic_list_widget->layout()->addWidget(xic_widget);
    for (int i=0; i < xic_sp_list.size(); i++) {
        qDebug() << "XicBox::setXic xic " << xic_sp_list[i].get();
        if (xic_sp_list[i].get() == nullptr) {
            throw new pappso::PappsoException("Error in XicBox::setXic:\n xic_sp_list[i].get() == nullptr");
        }
        qDebug() << "XicBox::setXic xic_sp_list[i].size() " << xic_sp_list[i].get()->size();
        ui->xic_widget->addXicSp(xic_sp_list[i]);
        QString isotope_name = QString("+%1").arg(_isotope_mass_list[i].get()->getIsotopeNumber());
        if (_isotope_mass_list[i].get()->getIsotopeRank() > 1) {
            isotope_name = QString("+%1 [%2]").arg(_isotope_mass_list[i].get()->getIsotopeNumber()).arg(_isotope_mass_list[i].get()->getIsotopeRank());
        }
        isotope_name.append(QString(" (%1%)").arg((int) (_isotope_mass_list[i].get()->getIntensityRatio()*100)));
        ui->xic_widget->setName(xic_sp_list[i].get(), isotope_name);

        if (_isotope_mass_list[i].get()->getIsotopeNumber() == 0) {
            for (const PeptideEvidence * peptide_evidence:_peptide_evidence_list) {
                qDebug() << "XicBox::setXic peptide_evidence " << peptide_evidence->getRetentionTime();
                ui->xic_widget->addMsMsEvent(xic_sp_list[i].get(), peptide_evidence->getRetentionTime());
            }
        }
    }
    qDebug() << "XicBox::setXic rescale" ;
    ui->xic_widget->rescale();

    /*
    if (_isotope_mass_list.size() > _xic_widget_list.size()) {
        emit loadXic(_p_peptide_evidence->getMsRunP(), _isotope_mass_list[_xic_widget_list.size()].get()->getMz(), _p_xic_window->getXicExtractPrecision(), XicExtractMethod::max);
    }
    */
    _xic_peak_match_list.clear();
    XicDetectionList xic_list;
    xic_list.setPeptideEvidenceList(_peptide_evidence_list);
    pappso::XicDetectionZivy zivy;
    zivy.setSmoothingHalfEdgeWindows(2);
    zivy.setDetectionThresholdOnMaxmin(3);
    zivy.setDetectionThresholdOnMinmax(3);
    zivy.setDetectionThresholdOnMaxmin(6000);
    zivy.setDetectionThresholdOnMinmax(5000);

    for (int i=0; i < xic_sp_list.size(); i++) {
        try {
            xic_list.clear();
            zivy.detect(*(xic_sp_list[i].get()), &xic_list);
            _xic_peak_match_list.push_back(xic_list.getMatchedPeak());
            ui->xic_widget->addXicPeakList(xic_sp_list[i].get(), xic_list.getXicPeakList());
        }
        catch (const pappso::ExceptionNotPossible & e) {
            qDebug() << e.qwhat();
        }
    }

    qDebug() << "XicBox::setXic plot" ;
    ui->xic_widget->plot();

    if (_isotope_ratio_graph_observed_intensity != nullptr) {
        _isotope_ratio_graph_observed_intensity->clearData();
    }
    _isotope_ratio_graph_observed_intensity = new QCPBars(ui->histo_widget->xAxis, ui->histo_widget->yAxis2);
    _isotope_ratio_graph_observed_intensity->setName("XIC area");
    //ui->histo_widget->yAxis2->setLabel("intensity");


    QVector<double> observed_intensity_data;
    QVector<QString> labels;

    QVector<double> ticks;
    double sum = 0;
    int i = 0;
    for (pappso::PeptideNaturalIsotopeAverageSp & natural_isotope_average: _isotope_mass_list) {
        sum += _xic_peak_match_list[i].getArea();
        observed_intensity_data << _xic_peak_match_list[i].getArea();
        QString isotope_name = QString("+%1").arg(_isotope_mass_list[i].get()->getIsotopeNumber());
        if (_isotope_mass_list[i].get()->getIsotopeRank() > 1) {
            isotope_name = QString("+%1 [%2]").arg(_isotope_mass_list[i].get()->getIsotopeNumber()).arg(_isotope_mass_list[i].get()->getIsotopeRank());
        }
        isotope_name.append(QString(" (%1%)").arg((int) (_isotope_mass_list[i].get()->getIntensityRatio()*100)));
        labels << isotope_name;
        ticks << i;
        i++;
    }

    _isotope_ratio_graph_observed_intensity->setPen(QPen(QColor("red")));
    //_graph_peak_surface_list.back()->setScatterStyle(QCPScatterStyle::ssDot);
    // observed_intensity->setBrush(QBrush(QColor(170, 255, 0, 0)));

    _isotope_ratio_graph_observed_intensity->setData(ticks, observed_intensity_data);

    ui->histo_widget->yAxis2->setVisible(true);
    ui->histo_widget->yAxis2->setRange(0, sum);
    ui->histo_widget->replot();
    qDebug() << "XicBox::setXic end";

}


void XicBox::setIsotopeMassList(std::vector<pappso::PeptideNaturalIsotopeAverageSp> isotope_mass_list) {
    _isotope_mass_list = isotope_mass_list;

    std::vector<pappso::mz> mass_list;
    for (pappso::PeptideNaturalIsotopeAverageSp & natural_isotope_average: _isotope_mass_list) {
        mass_list.push_back(natural_isotope_average.get()->getMz());
    }

    emit loadXic(_msrun_sp, mass_list, _p_xic_window->getXicExtractPrecision(), XicExtractMethod::max);

    //histogram
    QCPBars * theoretical_ratio = new QCPBars(ui->histo_widget->xAxis, ui->histo_widget->yAxis);
    theoretical_ratio->setName("theoretical ratio");
    ui->histo_widget->xAxis->setLabel("isotopes");
    //ui->histo_widget->yAxis->setLabel("th. ratio");

    QVector<double> theoretical_ratio_data;

    QVector<double> ticks;
    QVector<QString> labels;
    int i = 0;
    double sum = 0;
    for (pappso::PeptideNaturalIsotopeAverageSp & natural_isotope_average: _isotope_mass_list) {
        sum += natural_isotope_average.get()->getIntensityRatio();
        theoretical_ratio_data << natural_isotope_average.get()->getIntensityRatio();
        QString isotope_name = QString("+%1").arg(_isotope_mass_list[i].get()->getIsotopeNumber());
        if (_isotope_mass_list[i].get()->getIsotopeRank() > 1) {
            isotope_name = QString("+%1 [%2]").arg(_isotope_mass_list[i].get()->getIsotopeNumber()).arg(_isotope_mass_list[i].get()->getIsotopeRank());
        }
        isotope_name.append(QString(" (%1%)").arg((int) (_isotope_mass_list[i].get()->getIntensityRatio()*100)));
        labels << isotope_name;
        ticks << i;
        i++;
    }
    //QSharedPointer<QCPAxisTickerText> textTicker(new QCPAxisTickerText);
    //textTicker->addTicks(ticks, labels);
    //ui->histo_widget->xAxis->setTicker(textTicker);
    ui->histo_widget->xAxis->setTickLabelRotation(60);
    //ui->histo_widget->xAxis->setSubTicks(false);
    //ui->histo_widget->xAxis->setTickLength(0, 4);
    //ui->histo_widget->xAxis->setRange(0, 8);

    ui->histo_widget->xAxis->setAutoTickStep(false);   // <-- disable to use your own value

    ui->histo_widget->xAxis->setTickStep(1);


    theoretical_ratio->setData(ticks, theoretical_ratio_data);


    ui->histo_widget->yAxis->setRange(0, sum);
    ui->histo_widget->xAxis->setRange(-0.8, _isotope_mass_list.size());
    ui->histo_widget->replot();
}


void XicBox::reExtractXic() {
    qDebug() << "XicBox::reExtractXic begin";
    ui->xic_widget->clear();

    std::vector<pappso::mz> mass_list;
    for (pappso::PeptideNaturalIsotopeAverageSp & natural_isotope_average: _isotope_mass_list) {
        mass_list.push_back(natural_isotope_average.get()->getMz());
    }
    emit loadXic(_msrun_sp, mass_list, _p_xic_window->getXicExtractPrecision(), XicExtractMethod::max);
    qDebug() << "XicBox::reExtractXic end";
}
